"""
Functions for handling runtime function hooks.
"""

from collections import defaultdict
from functools import wraps
import inspect
import warnings
import sys


# global dict of hooks
# {class_name: { inst_id: {fname: [pre_hooks, post_hooks]}}}
# At class_name level, there might be a 'None' entry, which means that all instances of
# that class will have have the hooks specified in the 'None' dict.  Otherwise, entries
# there will be specific instance names.
# pre_hooks and post_hooks are lists of functions to call before or after the function named
# 'fname'.
_hooks = {}

# classes found here are known to contain no hooks within themselves or their ancestors
_hook_skip_classes = set()

# global switch that turns hook machinery on/off. But need it on for reporting system
use_hooks = True


def _reset_all_hooks():
    global _hooks, _hook_skip_classes
    _hooks = {}
    _hook_skip_classes = set()


def _setup_hooks(obj):
    """
    Wrap object's methods with a hook checking function if hooks are defined for any of them.

    Parameters
    ----------
    obj : object
        The object whose methods may be wrapped.
    """
    global _hooks, _hook_skip_classes

    # _setup_hooks should be called after 'obj' can return a valid name from _get_inst_id().
    # For example, in Problem, it can happen in __init__, but in Component and Group it shouldn't
    # happen until _setup_procs because that's the earliest point where the component/group has a
    # valid pathname.
    if use_hooks and obj.__class__ not in _hook_skip_classes:

        classes = inspect.getmro(obj.__class__)
        for c in classes:
            if c.__name__ in _hooks:
                classmeta = _hooks[c.__name__]
                break
        else:
            # didn't find any matching hooks for this class or any base class, so skip in future
            _hook_skip_classes.update(classes)
            return

        # any object where we register hooks must define the '_get_inst_id' method.
        ident = obj._get_inst_id()

        if ident in classmeta:
            instmeta = classmeta[ident]
        elif None in classmeta:  # ident of None applies to all instances of a class
            instmeta = classmeta[None]
        else:
            return

        for funcname in instmeta:
            method = getattr(obj, funcname, None)
            if method is not None:
                # if _hook_ attr is present, we've already wrapped this method
                if not hasattr(method, '_hook_'):
                    setattr(obj, funcname, _hook_decorator(method, obj, instmeta[funcname]))


def _run_hooks(hooks, inst):
    """
    Run the given list of hooks.

    Parameters
    ----------
    hooks : list
        List of hook data.
    inst : object
        Object instance to pass to hook functions.
    """
    for i, (hook, ncalls, ex, kwargs) in enumerate(hooks):
        if ncalls is None or ncalls > 0:
            hook(inst, **kwargs)
            if ex:
                sys.exit()
            if ncalls is not None:
                ncalls -= 1
                hooks[i] = (hook, ncalls, ex, kwargs)


def _hook_decorator(f, inst, hookmeta):
    """
    Wrap a method with pre and/or post hook checks.

    Parameters
    ----------
    f : method
        The method being wrapped.
    inst : object
        The instance that owns the method.
    hookmeta : dict
        A dict with information about the hooks.
    """
    pre_hooks, post_hooks = hookmeta

    # args and kwargs are arguments to the method that is being wrapped
    def execute_hooks(*args, **kwargs):
        _run_hooks(pre_hooks, inst)
        ret = f(*args, **kwargs)
        _run_hooks(post_hooks, inst)
        return ret

    execute_hooks._hook_ = True  # to prevent multiple decoration of same function

    return wraps(f)(execute_hooks)


def _get_hook_lists(class_name, inst_id, fname):
    """
    Retrieve the pre and post hook lists for the given class, instance, and function name.

    Parameters
    ----------
    class_name : str
        The name of the class owning the method where the hook will be applied.
    inst_id : str, optional
        The name of the instance owning the method where the hook will be applied.
    fname : str
        The name of the function where the pre and/or post hook will be applied.
    """
    global _hooks
    if class_name in _hooks:
        cmeta = _hooks[class_name]
    else:
        _hooks[class_name] = cmeta = {}
    if inst_id in cmeta:
        imeta = cmeta[inst_id]
    else:
        cmeta[inst_id] = imeta = {}
    if fname in imeta:
        return imeta[fname]
    imeta[fname] = [[], []]
    return imeta[fname]


def _register_hook(fname, class_name, inst_id=None, pre=None, post=None, ncalls=None, exit=False,
                   **kwargs):
    """
    Register a hook function.

    Parameters
    ----------
    hookfunc : function
        A function to execute in the designated hook location.
    fname : str
        The name of the function where the pre and/or post hook will be applied.
    class_name : str
        The name of the class owning the method where the hook will be applied.
    inst_id : str, optional
        The name of the instance owning the method where the hook will be applied.
    pre : function (None)
        If not None, this hook will run before the named function runs.
    post : function (None)
        If not None, this hook will run after the named function runs.
    ncalls : int or None
        Auto-remove the hook function after this many calls.  If None, never auto-remove.
        If both pre and post are registered, this will affect both.
    exit : bool
        If True, run sys.exit() after calling the hook function.  If post is registered, this
        affects only post, else it will affect pre.
    **kwargs : dict of keyword arguments
        Keyword arguments that will be passed to the hook function.
    """
    if pre is None and post is None:
        raise RuntimeError("In _register_hook you must specify pre or post.")

    pre_hooks, post_hooks = _get_hook_lists(class_name, inst_id, fname)
    if pre is not None and (ncalls is None or ncalls > 0):
        pre_hooks.append((pre, ncalls, exit and post is None, kwargs))
    if post is not None and (ncalls is None or ncalls > 0):
        post_hooks.append((post, ncalls, exit, kwargs))


def _remove_hook(to_remove, hooks, class_name, fname, hook_loc):
    """
    Remove a hook function.

    Parameters
    ----------
    to_remove : bool or function
        If True, all hook functions in 'hooks' will be removed.  If a function, any function
        in 'hooks' that matches will be removed.
    hooks : list
        List of (hook_func, ncalls, exit) tuples.
    class_name : str
        The name of the class owning the method where the hook will be removed.
    fname : str
        The name of the function where the hooks are located.
    hook_loc : str
        Either 'pre' or 'post', indicating the hooks run before or after respectively the
        function specified by fname.
    """
    if to_remove and hooks:
        if to_remove is True:
            hooks[:] = []
        else:
            for hook in hooks:
                p, _, _, _ = hook
                if p is to_remove:
                    hooks.remove(hook)
                    break
            else:
                raise RuntimeError(f"Couldn't find the given '{hook_loc}' function in the "
                                   f"{hook_loc} hooks for {class_name}.{fname}.")


def _unregister_hook(fname, class_name, inst_id=None, pre=True, post=True):
    """
    Unregister a hook function.

    By default, both pre and post hooks will be removed if they are present. To avoid
    removal of pre or post, you must set the pre or post arg to False.

    Parameters
    ----------
    fname : str
        The name of the function where the pre and/or post hook will be removed.
    class_name : str
        The name of the class owning the method where the hook will be removed.
    inst_id : str, optional
        The name of the instance owning the method where the hook will be removed.
    pre : bool or function, (True)
        If True, hooks that run before the named function runs will be removed. If a
        function then that function, if found, will be removed from the pre list, else
        an exception will be raised.
    post : bool or function, (True)
        If True, hooks that run after the named function runs will be removed. If a
        function then that function, if found, will be removed from the post list, else
        an exception will be raised.
    """
    try:
        hookdict = _hooks[class_name][inst_id]
    except KeyError:
        warnings.warn(f"No hooks found for class '{class_name}' and instance '{inst_id}'.")
        return

    if fname in hookdict:
        pre_hooks, post_hooks = hookdict[fname]
        _remove_hook(pre, pre_hooks, class_name, fname, 'pre')
        _remove_hook(post, post_hooks, class_name, fname, 'post')

        if not (pre_hooks or post_hooks):
            del hookdict[fname]
            if not hookdict:  # we just removed the last hook entry for this inst
                del _hooks[class_name][inst_id]
                if not _hooks[class_name]:  # removed last entry for this class
                    del _hooks[class_name]
    else:
        warnings.warn(f"No hook found for method '{fname}' for class '{class_name}' and instance "
                      f"'{inst_id}'.")



"""Unit Tests for the code that does automatic report generation"""
import unittest
import pathlib
import sys
import os
from io import StringIO

import openmdao.api as om
from openmdao.test_suite.components.paraboloid import Paraboloid
from openmdao.test_suite.components.sellar_feature import SellarMDA
import openmdao.core.problem
from openmdao.core.constants import _UNDEFINED
from openmdao.utils.assert_utils import assert_warning
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.reports_system import set_default_reports_dir, _reports_dir, register_report, \
    list_reports, clear_reports, run_n2_report, setup_default_reports, report_function
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.mpi import MPI
from openmdao.utils.tests.test_hooks import hooks_active
from openmdao.visualization.n2_viewer.n2_viewer import _default_n2_filename
from openmdao.visualization.scaling_viewer.scaling_report import _default_scaling_filename

try:
    from openmdao.vectors.petsc_vector import PETScVector
except ImportError:
    PETScVector = None

OPT, OPTIMIZER = set_pyoptsparse_opt('SLSQP')

if OPTIMIZER:
    from openmdao.drivers.pyoptsparse_driver import pyOptSparseDriver


@use_tempdirs
class TestReportsSystem(unittest.TestCase):

    def setUp(self):
        self.n2_filename = _default_n2_filename
        self.scaling_filename = _default_scaling_filename

        # set things to a known initial state for all the test runs
        openmdao.core.problem._problem_names = []  # need to reset these to simulate separate runs
        os.environ.pop('OPENMDAO_REPORTS', None)
        os.environ.pop('OPENMDAO_REPORTS_DIR', None)
        # We need to remove the TESTFLO_RUNNING environment variable for these tests to run.
        # The reports code checks to see if TESTFLO_RUNNING is set and will not do anything if set
        # But we need to remember whether it was set so we can restore it
        self.testflo_running = os.environ.pop('TESTFLO_RUNNING', None)
        clear_reports()
        set_default_reports_dir(_reports_dir)

        self.count = 0

    def tearDown(self):
        # restore what was there before running the test
        if self.testflo_running is not None:
            os.environ['TESTFLO_RUNNING'] = self.testflo_running

    def setup_and_run_simple_problem(self, driver=None, reports=_UNDEFINED, reports_dir=_UNDEFINED):
        prob = om.Problem(reports=reports, reports_dir=reports_dir)
        model = prob.model

        model.add_subsystem('p1', om.IndepVarComp('x', 0.0), promotes=['x'])
        model.add_subsystem('p2', om.IndepVarComp('y', 0.0), promotes=['y'])
        model.add_subsystem('comp', Paraboloid(), promotes=['x', 'y', 'f_xy'])

        model.add_design_var('x', lower=0.0, upper=1.0)
        model.add_design_var('y', lower=0.0, upper=1.0)
        model.add_objective('f_xy')

        if driver:
            prob.driver = driver

        else:
            prob.driver = om.ScipyOptimizeDriver()

        prob.setup(False)
        prob.run_driver()
        prob.cleanup()

        return prob

    def setup_and_run_model_with_subproblem(self, prob1_reports=_UNDEFINED,
                                            prob2_reports=_UNDEFINED):
        class _ProblemSolver(om.NonlinearRunOnce):

            def __init__(self, prob_name=None, reports=_UNDEFINED):
                super(_ProblemSolver, self).__init__()
                self.prob_name = prob_name
                self.reports = reports
                self._problem = None

            def solve(self):
                subprob = om.Problem(name=self.prob_name, reports=self.reports)
                self._problem = subprob
                subprob.model.add_subsystem('indep', om.IndepVarComp('x', 1.0))
                subprob.model.add_subsystem('comp', om.ExecComp('y=2*x'))
                subprob.model.connect('indep.x', 'comp.x')
                subprob.setup()
                subprob.run_model()

                return super().solve()

        prob = om.Problem(reports=prob1_reports)
        prob.model.add_subsystem('indep', om.IndepVarComp('x', 1.0))
        G = prob.model.add_subsystem('G', om.Group())
        G.add_subsystem('comp', om.ExecComp('y=2*x'))
        G.nonlinear_solver = _ProblemSolver(reports=prob2_reports)
        prob.model.connect('indep.x', 'G.comp.x')
        prob.setup()
        prob.run_model()  # need to do run_model in this test so sub problem is created

        probname = prob._name
        subprobname = G.nonlinear_solver._problem._name
        return probname, subprobname

    @hooks_active
    def test_report_generation_basic(self):
        setup_default_reports()
        prob = self.setup_and_run_simple_problem()

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertTrue(path.is_file(), f'The scaling report file, {str(path)}, was not found')

    @hooks_active
    @unittest.skipUnless(OPTIMIZER, "This test requires pyOptSparseDriver.")
    def test_report_generation_basic_pyoptsparse(self):
        # Just to try a different driver
        setup_default_reports()
        prob = self.setup_and_run_simple_problem(driver=pyOptSparseDriver(optimizer='SLSQP'))

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertTrue(path.is_file(), f'The scaling report file, {str(path)}, was not found')

    @hooks_active
    def test_report_generation_basic_doedriver(self):
        # Test a driver that does not generate scaling report
        setup_default_reports()
        prob = self.setup_and_run_simple_problem(driver=om.DOEDriver(om.PlackettBurmanGenerator()))

        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        # DOEDriver won't cause the creation of a scaling report
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_list_reports(self):
        setup_default_reports()  # So it sees the OPENMDAO_REPORTS var
        stdout = sys.stdout
        strout = StringIO()
        sys.stdout = strout
        try:
            list_reports()
        finally:
            sys.stdout = stdout

        output = strout.getvalue()
        self.assertTrue('N2 diagram' in output,
                        '"N2 diagram" expected in list_reports output but was not found')
        self.assertTrue('Driver scaling report' in output,
                        '"Driver scaling report" expected in list_reports output but was not found')

    @hooks_active
    def test_report_generation_no_reports_using_env_var(self):
        # test use of the OPENMDAO_REPORTS variable to turn off reporting
        os.environ['OPENMDAO_REPORTS'] = 'false'
        clear_reports()
        setup_default_reports()  # So it sees the OPENMDAO_REPORTS var

        prob = self.setup_and_run_simple_problem()

        # See if the report files exist and if they have the right names
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertFalse(path.is_file(),
                         f'The N2 report file, {str(path)} was found but should not have')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_selected_reports_using_env_var(self):
        # test use of the OPENMDAO_REPORTS variable to turn off selected reports
        os.environ['OPENMDAO_REPORTS'] = 'n2'
        clear_reports()
        setup_default_reports()  # So it sees the OPENMDAO_REPORTS var

        prob = self.setup_and_run_simple_problem()

        # See if the report files exist and if they have the right names
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_set_reports_dir_using_env_var(self):
        setup_default_reports()  # So it sees the OPENMDAO_REPORTS var
        # test use of setting a custom reports directory other than the default of "."
        custom_dir = 'custom_reports_dir'
        os.environ['OPENMDAO_REPORTS_DIR'] = custom_dir

        prob = self.setup_and_run_simple_problem()

        # See if the report files exist and if they have the right names
        reports_dir = custom_dir
        problem_reports_dir = pathlib.Path(reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertTrue(path.is_file(), f'The scaling report file, {str(path)}, was not found')

    @hooks_active
    def test_report_generation_user_defined_report(self):
        setup_default_reports()  # So it sees the OPENMDAO_REPORTS var
        user_report_filename = 'user_report.txt'

        # @report_function(user_report_filename)
        @report_function()
        def user_defined_report(prob, report_filepath):
            with open(report_filepath, "w") as f:
                f.write(f"Do some reporting on the Problem, {prob._name}\n")

        # register_report("User defined report", user_defined_report,
        #                 "user defined report description",
        #                 'Problem', 'setup', 'pre')
        register_report("User defined report", user_defined_report,
                        "user defined report description",
                        'Problem', 'setup', 'pre', user_report_filename)

        prob = self.setup_and_run_simple_problem()

        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(user_report_filename)
        self.assertTrue(path.is_file(), f'The user defined report file, {str(path)} was not found')

    @hooks_active
    def test_report_generation_various_locations(self):
        # the reports can be generated pre and post for setup, final_setup, and run_driver
        # check those all work
        setup_default_reports()  # So it sees the OPENMDAO_REPORTS var

        self.count = 0

        # A simple report
        user_report_filename = 'user_defined_{count}.txt'

        # @report_function(user_report_filename)
        @report_function()
        def user_defined_report(prob, report_filepath):
            report_filepath = report_filepath.format(count=self.count)
            with open(report_filepath, "w") as f:
                f.write(f"Do some reporting on the Problem, {prob._name}\n")
            self.count += 1

        for method in ['setup', 'final_setup', 'run_driver']:
            for pre_or_post in ['pre', 'post']:
                register_report(f"User defined report {method} {pre_or_post}", user_defined_report,
                                "user defined report", 'Problem', method, pre_or_post,
                                user_report_filename)

        prob = self.setup_and_run_simple_problem()

        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        self.count = 0
        for _ in ['setup', 'final_setup', 'run_driver']:
            for _ in ['pre', 'post']:
                user_report_filename = f"user_defined_{self.count}.txt"
                path = pathlib.Path(problem_reports_dir).joinpath(user_report_filename)
                self.assertTrue(path.is_file(),
                                f'The user defined report file, {str(path)} was not found')
                self.count += 1

    @hooks_active
    def test_report_generation_multiple_problems(self):
        setup_default_reports()  # So it sees the OPENMDAO_REPORTS var
        probname, subprobname = self.setup_and_run_model_with_subproblem()

        # The multiple problem code only runs model so no scaling reports to look for
        for problem_name in [probname, subprobname]:
            problem_reports_dir = pathlib.Path(_reports_dir).joinpath(f'{problem_name}')
            path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
            self.assertTrue(path.is_file(), f'N2 report file, {str(path)} was not found')

    @hooks_active
    def test_report_generation_multiple_problems_report_specific_problem(self):
        # test the ability to register a report with a specific Problem name rather
        #   than have the report run for all Problems

        # to simplify things, just do n2.
        clear_reports()
        register_report("n2_report", run_n2_report, 'N2 diagram', 'Problem', 'final_setup', 'post',
                        self.n2_filename,
                        inst_id='problem2')

        probname, subprobname = self.setup_and_run_model_with_subproblem()

        # The multiple problem code only runs model so no scaling reports to look for
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(f'{subprobname}')
        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        # for the subproblem named problem2, there should be a report but not for problem1 since
        #    we specifically asked for just the instance of problem2
        self.assertTrue(path.is_file(), f'The n2 report file, {str(path)} was not found')

        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(f'{probname}')
        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertFalse(path.is_file(),
                         f'The N2 report file, {str(path)} was found but should not have')

    @hooks_active
    def test_report_generation_test_TESTFLO_RUNNING(self):
        # need to do this here again even though it is done in setup, because otherwise
        # setup_default_reports won't see environment variable, TESTFLO_RUNNING
        os.environ['TESTFLO_RUNNING'] = 'true'
        clear_reports()
        setup_default_reports()

        prob = self.setup_and_run_simple_problem()

        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertFalse(path.is_file(),
                         f'The N2 report file, {str(path)} was found but should not have')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_basic_problem_reports_argument_false(self):
        setup_default_reports()

        prob = self.setup_and_run_simple_problem(reports=False)

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)
        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertFalse(path.is_file(),
                         f'The N2 report file, {str(path)} was found but should not have')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_basic_problem_reports_argument_none(self):
        setup_default_reports()

        prob = self.setup_and_run_simple_problem(reports=None)

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertFalse(path.is_file(),
                         f'The N2 report file, {str(path)} was found but should not have')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_basic_problem_reports_argument_n2_only(self):
        setup_default_reports()
        prob = self.setup_and_run_simple_problem(reports='n2')

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_basic_problem_reports_argument_n2_and_scaling(self):
        setup_default_reports()
        prob = self.setup_and_run_simple_problem(reports='n2,scaling')

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertTrue(path.is_file(), f'The scaling report file, {str(path)} was not found')

    @hooks_active
    def test_report_generation_problem_reports_argument_multiple_problems(self):
        setup_default_reports()
        _, _ = self.setup_and_run_model_with_subproblem(prob2_reports=None)

        # Only problem1 reports should have been generated

        # The multiple problem code only runs model so no scaling reports to look for
        problem_name = 'problem1'
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(f'{problem_name}')
        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The problem1 N2 report file, {str(path)} was not found')

        problem_name = 'problem2'
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(f'{problem_name}')
        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertFalse(path.is_file(),
                         f'The problem2 n2 report file, {str(path)}, was found but should not have')

    @hooks_active
    def test_report_generation_basic_problem_reports_dir_argument(self):
        setup_default_reports()

        custom_reports_dir = 'user_dir'

        prob = self.setup_and_run_simple_problem(reports=False, reports_dir=custom_reports_dir)

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(custom_reports_dir).joinpath(prob._name)
        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertFalse(path.is_file(),
                         f'The N2 report file, {str(path)} was found but should not have')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertFalse(path.is_file(),
                         f'The scaling report file, {str(path)}, was found but should not have')


@use_tempdirs
@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
class TestReportsSystemMPI(unittest.TestCase):
    N_PROCS = 2

    def setUp(self):
        self.n2_filename = _default_n2_filename
        self.scaling_filename = _default_scaling_filename

        # set things to a known initial state for all the test runs
        openmdao.core.problem._problem_names = []  # need to reset these to simulate separate runs

        os.environ.pop('OPENMDAO_REPORTS', None)
        os.environ.pop('OPENMDAO_REPORTS_DIR', None)
        # We need to remove the TESTFLO_RUNNING environment variable for these tests to run.
        # The reports code checks to see if TESTFLO_RUNNING is set and will not do anything
        # if it is set.
        # But we need to remember whether it was set so we can restore it
        self.testflo_running = os.environ.pop('TESTFLO_RUNNING', None)
        clear_reports()
        set_default_reports_dir(_reports_dir)

        self.count = 0  # used to keep a count of reports generated

    def tearDown(self):
        # restore what was there before running the test
        if self.testflo_running is not None:
            os.environ['TESTFLO_RUNNING'] = self.testflo_running

    @hooks_active
    def test_reports_system_mpi_basic(self):  # example taken from TestScipyOptimizeDriverMPI
        setup_default_reports()

        prob = om.Problem()
        prob.model = SellarMDA()
        prob.driver = om.ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-8)

        prob.model.add_design_var('x', lower=0, upper=10)
        prob.model.add_design_var('z', lower=0, upper=10)
        prob.model.add_objective('obj')
        prob.model.add_constraint('con1', upper=0)
        prob.model.add_constraint('con2', upper=0)

        prob.setup()
        prob.set_solver_print(level=0)

        prob.run_driver()

        # get the path to the problem subdirectory
        problem_reports_dir = pathlib.Path(_reports_dir).joinpath(prob._name)

        path = pathlib.Path(problem_reports_dir).joinpath(self.n2_filename)
        self.assertTrue(path.is_file(), f'The N2 report file, {str(path)} was not found')
        path = pathlib.Path(problem_reports_dir).joinpath(self.scaling_filename)
        self.assertTrue(path.is_file(), f'The scaling report file, {str(path)}, was not found')


if __name__ == '__main__':
    unittest.main()



"""
Utility functions related to the reporting system which generates reports by default for all runs.
"""

from collections import namedtuple, defaultdict
from functools import wraps
import pathlib
import sys
import os

from openmdao.utils.mpi import MPI
from openmdao.utils.hooks import _register_hook, _unregister_hook
from openmdao.visualization.n2_viewer.n2_viewer import n2, _default_n2_filename
from openmdao.visualization.scaling_viewer.scaling_report import _default_scaling_filename
from openmdao.core.constants import _UNDEFINED
from openmdao.core.problem import Problem
from openmdao.core.driver import Driver

# Keeping track of the registered reports
_Report = namedtuple('Report', 'func desc class_name inst_id method pre_or_post report_filename')
_reports_registry = {}

_reports_dir = './reports'  # the default location for the reports


def _is_rank_0(prob):
    # Utility function to determine if MPI and on rank0 or not on MPI at all
    return not MPI or prob.comm.rank == 0


def report_function():
    """
    Decorate report functions. Handles getting the file path to where the report is written.

    Returns
    -------
    function
        The wrapper function.
    """
    def decorate(f):
        @wraps(f)
        def _wrapper(inst, **kwargs):
            if isinstance(inst, Problem):
                prob = inst
            elif isinstance(inst, Driver):
                prob = inst._problem()
            else:
                raise ValueError("User defined reports currently can only be registered "
                                 "on Problems and Drivers")

            problem_reports_dirpath = get_reports_dir(prob)
            if _is_rank_0(prob):
                pathlib.Path(problem_reports_dirpath).mkdir(parents=True, exist_ok=True)

            report_filename = kwargs['report_filename']
            user_defined_report_filepath = \
                str(pathlib.Path(problem_reports_dirpath).joinpath(report_filename))

            f(inst, user_defined_report_filepath)
        return _wrapper
    return decorate


def register_report(name, func, desc, class_name, method, pre_or_post, report_filename,
                    inst_id=None):
    """
    Register a report with the reporting system.

    Parameters
    ----------
    name : str
        Name of report. Report names must be unique across all reports.
    func : function
        A function to do the reporting. Expects the first argument to be an instance of class_name.
    desc : str
        A description of the report.
    class_name : str
        The name of the class owning the method where the report will be run.
    method : str
        In which method of class_name should this be run.
    pre_or_post : str
        Valid values are 'pre' and 'post'. Indicates when to run the report in the method.
    report_filename : str
        Name of file to use when saving the report.
    inst_id : str or None
        Either the instance ID of an OpenMDAO object (e.g. Problem, Driver) or None.
        If None, then this report will be run for all objects of type class_name.
    """
    global _reports_registry

    report = _Report(func, desc, class_name, inst_id, method, pre_or_post, report_filename)
    if name in _reports_registry:
        raise ValueError(f"A report with the name {name} already exists")
    _reports_registry[name] = report

    if pre_or_post == 'pre':
        _register_hook(method, class_name, pre=func, inst_id=inst_id,
                       report_filename=report_filename)
    elif pre_or_post == 'post':
        _register_hook(method, class_name, post=func, inst_id=inst_id,
                       report_filename=report_filename)
    else:
        raise ValueError(
            f"The argument 'pre_or_post' can only have values of 'pre' or 'post', but {pre_or_post}"
            " was given")
    return


def list_reports(out_stream=None):
    """
    Write table of information about reports currently registered in the reporting system.

    Parameters
    ----------
    out_stream : file-like object
        Where to send report info.
    """
    global _reports_registry

    if not out_stream:
        out_stream = sys.stdout

    column_names = ['name', 'desc', 'class_name', 'inst_id', 'method', 'pre_or_post', 'func']
    column_widths = {}
    # Determine the column widths of the data fields by finding the max width for all rows
    # First for the headers
    for column_name in column_names:
        column_widths[column_name] = len(column_name)

    # Now for the values
    for name, report in _reports_registry.items():
        for column_name in column_names:
            if column_name == 'name':
                val = name
            else:
                val = getattr(report, column_name)
                if column_name == 'func':
                    val = val.__name__
                else:
                    val = str(val)
            column_widths[column_name] = max(column_widths[column_name], len(val))

    out_stream.write("\nHere are the reports available:\n\n")

    column_header = ''
    column_dashes = ''
    column_spacing = 2
    for i, column_name in enumerate(column_names):
        column_header += '{:{width}}'.format(column_name, width=column_widths[column_name])
        column_dashes += column_widths[column_name] * '-'
        if i < len(column_names) - 1:
            column_header += column_spacing * ' '
            column_dashes += column_spacing * ' '

    out_stream.write('\n')
    out_stream.write(column_header + '\n')
    out_stream.write(column_dashes + '\n')

    for name, report in _reports_registry.items():
        report_info = ''
        for i, column_name in enumerate(column_names):
            if column_name == 'name':
                val = name
            else:
                val = getattr(report, column_name)
                if column_name == 'func':
                    val = val.__name__
                else:
                    val = str(val)
            val_formatted = f"{val:<{column_widths[column_name]}}"
            report_info += val_formatted
            if i < len(column_names) - 1:
                report_info += column_spacing * ' '

        out_stream.write(report_info + '\n')

    out_stream.write('\n')


def set_default_reports_dir(reports_dir_path):
    """
    Set the path to where the reports should go. By default, they go into the current directory.

    Parameters
    ----------
    reports_dir_path : str
        Path to where the report directories should go.
    """
    global _reports_dir
    _reports_dir = reports_dir_path


def get_reports_dir(prob):
    """
    Get the path to the directory where the report files should go.

    Parameters
    ----------
    prob : OpenMDAO Problem object
        The report will be run on this Problem.

    Returns
    -------
    str
        The path to the directory where reports should be written.
    """
    if prob._reports_dir is not _UNDEFINED:
        reports_dir = prob._reports_dir
    else:
        reports_dir = os.environ.get('OPENMDAO_REPORTS_DIR', _reports_dir)

    # problem_reports_dirname = f'{prob._name}_reports'
    problem_reports_dirname = f'{prob._name}'
    problem_reports_dirpath = pathlib.Path(reports_dir).joinpath(problem_reports_dirname)

    return problem_reports_dirpath


def clear_reports():
    """
    Clear all of the reports from the registry.
    """
    global _reports_registry

    # need to remove the hooks
    for name, report in _reports_registry.items():
        if getattr(report, 'pre_or_post') == "pre":
            _unregister_hook(getattr(report, 'method'), getattr(report, 'class_name'),
                             inst_id=getattr(report, 'inst_id'), pre=getattr(report, 'func'))
        else:
            _unregister_hook(getattr(report, 'method'), getattr(report, 'class_name'),
                             inst_id=getattr(report, 'inst_id'), post=getattr(report, 'func'))
    _reports_registry = {}


def _should_report_run(reports, report_name):
    # Utility function that checks the _reports attribute on Problem
    #   to determine whether the report named "report_name" should be run
    if isinstance(reports, str):
        reports_on = reports.split(',')
        if report_name not in reports_on:
            return False
    elif isinstance(reports, bool):
        if not reports:
            return False
    elif reports is None:
        return False

    return True


# N2 report definition
# Need to create these closures so that functions can keep track of how many times they have
# been called per Problem (or Driver). In the case of the n2, it is Problem
def _run_n2_report_enclosing():
    def run_n2_report_inner(prob, report_filename=None):

        if not _should_report_run(prob._reports, 'n2'):
            return

        run_n2_report_inner.calls[prob] += 1
        if run_n2_report_inner.calls[prob] > 1:  # Only do the report once per Problem
            return

        problem_reports_dirpath = get_reports_dir(prob)
        if _is_rank_0(prob):
            pathlib.Path(problem_reports_dirpath).mkdir(parents=True, exist_ok=True)

        n2_filepath = str(pathlib.Path(problem_reports_dirpath).joinpath(report_filename))
        try:
            n2(prob, show_browser=False, outfile=n2_filepath, display_in_notebook=False)
        except RuntimeError as err:
            # We ignore this error
            if str(err) != "Can't compute total derivatives unless " \
                           "both 'of' or 'wrt' variables have been specified.":
                raise err

    run_n2_report_inner.calls = defaultdict(int)
    return run_n2_report_inner


run_n2_report = _run_n2_report_enclosing()


# scaling report definition
def _run_scaling_report_enclosing():
    # def run_scaling_report_inner(driver):
    def run_scaling_report_inner(driver, report_filename=None):

        prob = driver._problem()

        if not _should_report_run(prob._reports, 'scaling'):
            return

        run_scaling_report_inner.calls[driver] += 1
        if run_scaling_report_inner.calls[driver] > 1:
            return

        problem_reports_dirpath = get_reports_dir(prob)

        scaling_filepath = str(
            pathlib.Path(problem_reports_dirpath).joinpath(report_filename))
        if _is_rank_0(prob):
            pathlib.Path(problem_reports_dirpath).mkdir(parents=True, exist_ok=True)

        try:
            prob.driver.scaling_report(outfile=scaling_filepath, show_browser=False)

        # Need to handle the coloring and scaling reports which can fail in this way
        #   because total Jacobian can't be computed
        except RuntimeError as err:
            if str(err) != "Can't compute total derivatives unless " \
                           "both 'of' or 'wrt' variables have been specified.":
                raise err

    run_scaling_report_inner.calls = defaultdict(int)
    return run_scaling_report_inner


run_scaling_report = _run_scaling_report_enclosing()

_default_reports = {
    'n2': (run_n2_report, 'N2 diagram', 'Problem', 'final_setup', 'post', _default_n2_filename,
           None),
    'scaling': (run_scaling_report, 'Driver scaling report', 'Driver', '_compute_totals', 'post',
                _default_scaling_filename, None)
}


def setup_default_reports():
    """
    Set up the default reports for all OpenMDAO runs.
    """
    if 'TESTFLO_RUNNING' in os.environ:
        return

    if 'OPENMDAO_REPORTS' in os.environ:
        if os.environ['OPENMDAO_REPORTS'] in ['0', 'false', 'off', "none"]:
            return  # do not do any reports

        if os.environ['OPENMDAO_REPORTS'] in ['1', 'true', 'on', "all"]:
            reports_on = _default_reports.keys()
        else:
            reports_on = os.environ['OPENMDAO_REPORTS'].split(',')
    else:  # if no env set, all reports are on
        reports_on = _default_reports.keys()

    for report_name, report_info in _default_reports.items():
        if report_name in reports_on:
            register_report(report_name, *report_info)


setup_default_reports()



"""
Define the HtmlPreprocessor class to generate a single HTML file from many source files.
"""
import base64
import re
import zlib
import json
from pathlib import Path


class HtmlPreprocessor():
    """
    Recursively substitute and insert source files to produce a single HTML file.

    Source files contain text with directives in the form: <<hpp_directive arg [flag1] [flag2]>>

    Parameters
    ----------
    start_filename : str
        The file to begin processing from.
    output_filename : str
        The path to the newly merged HTML file.
    allow_overwrite : bool
        If true, overwrite the output file if it exists.
    var_dict : dict
        Dictionary of variable names and values referenced by hpp_pyvar directives.
    json_dumps_default : function
        Passed to json.dumps() as the "default" parameter that gets
        called for objects that can't be serialized.
    verbose : bool
        If True, print some status messages to stdout.

    Attributes
    ----------
    _start_filename : str
        The path to the file to begin processing from, normally an HTML file.
    _start_path : Path
        Path object used to determine relative paths for subsequent directives.
    _start_dirname : str
        The absolute path of the directory containing the start_file.
    _output_filename : str
        The path to the newly merged HTML file.
    _var_dict : dict
        Dictionary of variable names and values referenced by hpp_pyvar directives.
    _allow_overwrite : bool
        If true, overwrite the output file if it exists.
    _verbose : bool
        If True, print some status messages to stdout.
    _loaded_filenames : list
        List of filenames loaded so far, to determine if a file has been already loaded.
    _json_dumps_default : function
        Passed to json.dumps() as the "default" parameter that gets
        called for objects that can't be serialized.

    Notes
    -----
    Recognized Directives:

    :code:`<<hpp_insert path/to/file [compress] [dup]>>`
        Paste :code:`path/to/file` verbatim into the text.

    :code:`<<hpp_script path/to/script [dup]>>`
        Insert :code:`path/to/script` between script tags.

    :code:`<<hpp_style path/to/css [dup]>>`
        Insert :code:`path/to/css` between style tags.

    :code:`<<hpp_bin2b64 path/to/file [dup]>>`
        Convert a binary file to a b64 string and insert it.

    :code:`<<hpp_pyvar variable_name [compress]>>`
        Insert the string value of the named Python variable. If the referenced
        variable is non-primitive, it's converted to JSON.

    Flags:

    :code:`compress` :
        The replacement content will be compressed and converted to
        a base64 string. It's up to the JavaScript code to decode and uncompress it.
    :code:`dup` :
        If a file has already been included once, it will be ignored on subsequent inclusions
        unless the dup flag is used.

    Commented directives (:code:`//`, :code:`/* */`, or :code:`<!-- -->`) will replace the
    entire comment. When a directive is commented, it can only be on a single line or the
    comment-ending characters will not be replaced.

    All paths in the directives are relative to the directory that the start file
    is located in unless it is absolute.

    Nothing is written until every directive has been successfully processed.
    """

    def __init__(self, start_filename: str, output_filename: str, allow_overwrite=False,
                 var_dict: dict = None, json_dumps_default=None, verbose=False):
        """
        Configure the preprocessor and validate file paths.
        """
        self._start_path = Path(start_filename)
        if self._start_path.is_file() is False:
            raise FileNotFoundError(f"Error: {start_filename} not found")

        output_path = Path(output_filename)
        if output_path.is_file() and not allow_overwrite:
            raise FileExistsError(f"Error: {output_filename} already exists")

        self._start_filename = start_filename
        self._start_dirname = self._start_path.resolve().parent
        self._output_filename = output_filename
        self._allow_overwrite = allow_overwrite
        self._var_dict = var_dict
        self._json_dumps_default = json_dumps_default
        self._verbose = verbose

        # Keep track of filenames already loaded, to make sure
        # we don't unintentionally include the exact same file twice.
        self._loaded_filenames = []

        self.msg("HtmlProcessor object created.")

    def load_file(self, filename: str, rlvl=0, binary=False, allow_dup=False) -> str:
        """
        Open and read the specified text file.

        Parameters
        ----------
        filename : str
            The path to the text file to read.
        rlvl : int
            Recursion level to help with indentation when verbose is enabled.
        binary : bool
            True if the file is to be opened in binary mode and converted to a base64 str.
        allow_dup : bool
            If False, return an empty string for a filename that's been previously loaded.

        Returns
        -------
        str
            The complete contents of the file.
        """
        path = Path(filename)
        pathname = filename if path.is_absolute() else self._start_dirname / filename

        if pathname in self._loaded_filenames and not allow_dup:
            self.msg(f"Ignoring previously-loaded file {filename}.", rlvl)
            return ""

        self._loaded_filenames.append(pathname)
        self.msg(f"Loading file {pathname}.", rlvl)

        if binary:
            with open(pathname, 'rb') as f:
                file_contents = str(base64.b64encode(f.read()).decode('UTF-8'))
        else:
            with open(pathname, 'r', encoding='UTF-8') as f:
                file_contents = str(f.read())

        return file_contents

    def msg(self, msg: str, rlvl=0):
        """
        Print a message to stdout if self.verbose is True.

        Parameters
        ----------
        msg : str
            The message to print.
        rlvl : int
            Recursion level to help with indentation when verbose is enabled.
        """
        if self._verbose:
            print(rlvl * '--' + msg)

    def parse_contents(self, contents: str, rlvl=0) -> str:
        """
        Find the preprocessor directives in the file and replace them with the desired content.

        Will recurse if directives are also found in the new content.

        Parameters
        ----------
        contents : str
            The contents of a preloaded text file.
        rlvl : int
            Recursion level to help with indentation when verbose is enabled.

        Returns
        -------
        str
            The complete contents represented as a base64 string.
        """
        # Find all possible keywords:
        keyword_regex = \
            r'(//|/\*|<\!--)?\s*<<\s*hpp_(\S+)\s+(\S+)(\s+compress|\s+dup)*\s*>>(\*/|-->)?'
        matches = re.finditer(keyword_regex, contents)
        rlvl += 1
        new_content = None

        for found_directive in matches:

            full_match = found_directive.group(0)
            comment_start = found_directive.group(1)
            keyword = found_directive.group(2)
            arg = found_directive.group(3)

            flags = {'compress': False, 'dup': False}
            if found_directive.group(4) is not None:
                flags['compress'] = True if 'compress' in found_directive.group(4) else False
                flags['dup'] = True if 'dup' in found_directive.group(4) else False

            do_compress = False  # Change below with directives where it's allowed

            self.msg(f"Handling {keyword} directive.", rlvl)

            if keyword == 'insert':
                # Recursively insert a plain text file which may also have hpp directives
                new_content = self.parse_contents(self.load_file(arg, rlvl=rlvl,
                                                  allow_dup=flags['dup']), rlvl)

                if new_content != "":
                    do_compress = True if flags['compress'] else False

            elif keyword == 'script':
                # Recursively insert a JavaScript file which may also have hpp directives
                new_content = self.parse_contents(self.load_file(arg, rlvl=rlvl,
                                                  allow_dup=flags['dup']), rlvl)

                if new_content != "":
                    new_content = f'<script type="text/javascript">\n{new_content}\n</script>'

            elif keyword == 'style':
                # Recursively insert a CSS file which may also have hpp directives
                new_content = '<style type="text/css">\n' + \
                              self.parse_contents(self.load_file(arg, rlvl=rlvl,
                                                  allow_dup=flags['dup']), rlvl) + f'\n</style>'

            elif keyword == 'bin2b64':
                new_content = self.load_file(arg, binary=True, allow_dup=flags['dup'], rlvl=rlvl)

            elif keyword == 'pyvar':
                if arg in self._var_dict:
                    val = self._var_dict[arg]
                    if type(val) in (str, bool, int, float):
                        # Use string representations of primitive types
                        new_content = str(self._var_dict[arg])
                        do_compress = True if flags['compress'] else False
                    else:
                        raw_data = json.dumps(val, default=self._json_dumps_default)
                        if flags['compress']:
                            self.msg("Compressing content.", rlvl)
                            compressed_content = zlib.compress(raw_data.encode('UTF-8'))
                            new_content = str(base64.b64encode(compressed_content).decode("UTF-8"))
                        else:
                            new_content = raw_data

                else:
                    raise ValueError(f"Can't substitute for undefined variable {arg}")

            else:
                # Bad keyword
                raise ValueError(f"Unknown HTML preprocessor directive hpp_{keyword}")

            if do_compress:
                self.msg("Compressing new content.", rlvl)
                new_content = str(base64.b64encode(zlib.compress(new_content)).decode("UTF-8"))

            if new_content is not None:
                self.msg(f"Replacing directive '{full_match}' with new content.", rlvl)
                # contents = re.sub(full_match, new_content, contents)
                contents = contents.replace(full_match, new_content)

        return contents

    def run(self) -> None:
        """
        Initialize the preprocessor, then save the result as a new file.
        """
        new_html_content = self.parse_contents(self.load_file(self._start_filename))

        path = Path(self._output_filename)
        if path.is_file() and not self._allow_overwrite:
            raise FileExistsError(f"Error: {self._output_filename} already exists")

        output_file = open(self._output_filename, "w", encoding='UTF-8')
        output_file.write(new_html_content)
        output_file.close()



import unittest
import tracemalloc
import gc
import openmdao.api as om
from openmdao.test_suite.components.sellar_feature import SellarMDA

# Compare memory usage after running setup 10 times to
# running setup 100 times
ITERS = [ 10, 100 ]

# The code with the leak problem would report a difference of memory
# used between 10 and 100 iter runs of 36.9MiB. After fixing, it was
# 0.2 KiB.
MAX_MEM_DIFF_KB = 200

class TestSetupMemLeak(unittest.TestCase):
    """ Test for memory leaks when calling setup() multiple times """

    def test_setup_memleak(self):
        # Record the memory still in use after each run
        mem_used = []

        # Setup of the Sellar poblem
        prob = om.Problem()

        tracemalloc.start()

        for memtest in ITERS:
            prob.model = SellarMDA()
            snapshots = []

            for i in range(memtest):
                prob.setup(check=False) # called here causes memory leak   
                prob.run_driver()
                totals = prob.compute_totals("z", "x")
                del totals

                # Force garbage collection now instead of waiting for an
                # optimal/arbitrary point since we're tracking memory usage
                gc.collect()
                snapshots.append(tracemalloc.take_snapshot())

            top_stats = snapshots[memtest - 1].compare_to(snapshots[0], 'lineno')

            total_mem = 0
            for stat in top_stats:
                tb_str = str(stat.traceback)
                # Ignore memory allocations by tracemalloc itself, which throw things off:
                if "/tracemalloc.py:" not in tb_str:
                    total_mem += stat.size
            
            mem_used.append(total_mem)

        tracemalloc.stop()    
        mem_diff = (mem_used[1] - mem_used[0])/1024

        self.assertLess(mem_diff, MAX_MEM_DIFF_KB,
            "Memory leak in setup(): %.1f KiB difference between %d and %d iter runs" %
                (mem_diff, ITERS[0], ITERS[1]))

if __name__ == '__main__':
    unittest.main()